import java.util.Vector;
import java.util.Random;
import java.lang.Math.*;



/**
 * Class which implements the Location Diagram. This is the directed acyclic graph used during the Delaunay diagram construction
 */
public class LocationStructure{

    private Vector headList;
    private Vector leafList;

    private int recN=0;
    private int probN =0;

    private double distCaract;

    private Random generator = new Random();

    /**
     * Constructor of LocationStructure
     * @param triangleList Vector of Triangle objects used to initialized the LocationStructure
     */
    public LocationStructure(Vector triangleList){

	headList = new Vector();
	
	int ts = triangleList.size();
	for(int i=0;i<ts;i++){
	    Triangle tr = (Triangle)(triangleList.elementAt(i));
	    GraphNode node = new GraphNode(tr,this);
	    headList.addElement(node);
	}
	
	leafList = (Vector)(headList.clone());

    }


    /**
     * Generates the location structure graph 
     * @param numb the desired number of points for the associated Delauney triangulation
     */
    public void build(int numb){

	distCaract = 0.3*Math.sqrt(4.0*Math.PI/((double)numb));

	boolean inserted = false;
	int nb = 0;

	while (nb != numb){
	    Point tempP = this.pointGeneration(nb);
	    inserted  = insert(tempP);
	    nb++;
	}

    }


    /**
     * Manages the insertion of a new point in the associated Delauney graph
     * @param pr the Point to be inserted
     */
    public boolean insert(Point pr){

	GraphNode n  =findNode(pr);
	
	Triangle t = n.getTriangle();

	Point p1 = t.getS1();
	Point p2 = t.getS2();
	Point p3 = t.getS3();

	if ((pr.eucliDist(p1)<distCaract) || (pr.eucliDist(p2)<distCaract) ||
	    (pr.eucliDist(p3)<distCaract)){
	    return false;
	}
	else{

	    Triangle t1 = new Triangle(pr,p1,p2);
	    Triangle t2 = new Triangle(pr,p2,p3);
	    Triangle t3 = new Triangle(pr,p1,p3);

	    GraphNode n1 = new GraphNode(t1,this);
	    GraphNode n2 = new GraphNode(t2,this);
	    GraphNode n3 = new GraphNode(t3,this);

	    // the three new nodes are added
	    n.addChild( n1 );
	    n.addChild( n2 );
	    n.addChild( n3 );

	    // the leaf list is updated
	    if( !(leafList.removeElement(n)) ){
		System.err.println("#### pb dans la methode insert: pas d'element a enlever");
	    }
	    leafList.addElement(n1);
	    leafList.addElement(n2);
	    leafList.addElement(n3);
	

	    recN=0;
	    legalize(n1);
	    recN=0;
	    legalize(n2);
	    recN=0;
	    legalize(n3);
	    return true;
	}
    }

    
    /**
     * Ensures the legalization of a newly created node
     * @param node the GraphNode object to be legalized
     */
    private void legalize(GraphNode node){

	if (!leafList.contains(node)){
	    System.out.println("#### element a legalizer non contenu !");
	}
	
	recN++;

	Triangle tr = node.getTriangle();

	GraphNode nk = opposite(node);
	Triangle tk = nk.getTriangle();
	Point pk = nk.difference(node);

	if(tr.circumContains(pk)){

	    // Warning : pk has to be at the first position
	    GraphNode n1 = new GraphNode( new Triangle(tr.getS1(),pk, tr.getS2() ), this );
	    GraphNode n2 = new GraphNode( new Triangle(tr.getS1(),pk, tr.getS3() ), this );

	    node.addChild(n1);
	    node.addChild(n2);
	    
	    nk.addChild(n1);
	    nk.addChild(n2);

	    if( !(leafList.removeElement(node)) ){
		System.err.println("##### pb " + probN + ": feuille -node- a enlever non trouvee dans legalize." 
				   + "Appel recursif : " + recN );
		probN++;
		node.getTriangle().printCoordinates();
		if (leafList.contains(node)){
		    System.out.println("#### element pourtant contenu !");
		}
	    }
	    if( !(leafList.removeElement(nk)) ){
		System.err.println("##### pb: feuille -nk- a enlever non trouvee dans legalize ");
	    }

	    leafList.addElement(n1);
	    leafList.addElement(n2);

	    if (!leafList.contains(n1)){
		System.out.println("point -N1- venant d'etre ajoute est non contenu !");
	    }
	    if (!leafList.contains(n2)){
		System.out.println("point -N2- venant d'etre ajoute est non contenu !");
	    }

	    legalize(n1);
	    legalize(n2);
	}
    }
    



    /**
     * Finds in the leaf list the node which owns the other triangle with the edge (node->S2|node->S3)
     * NB : if it is not found, that means something is wrong.
     * @param node the reference node
     * @return tempno the found node
     */
    private GraphNode opposite(GraphNode node){

	Triangle tr = node.getTriangle();

	GraphNode tempno = null;
	Triangle temptr;

	int test =0;
	
	int ls = leafList.size();
	for(int i=0;i<ls;i++){

	    tempno = (GraphNode)(leafList.elementAt(i));
	    temptr = tempno.getTriangle();

	    // we check if the node "tno" is different from the argument
	    // and if the associated triangle caontains the common edge 
	    if( (tempno != node) && (temptr.owns(tr.getS2(),tr.getS3())) ){
		test=1;
		break;
	    }
	    	    
	}
	if( test == 0 ){
	    System.err.println("#### pb " + probN + " : pas de triangle oppose trouve");
	    probN++;
	}

	return tempno;

    }


    
    /**
     * Returns the node associated with the triangle of the Delaunay triangulation in which the point pr is located
     * @param pr the Point that is to be located in the Delaunay triangulation
     * @return the node owning the triangle in which the Point pr is located
     */
    private GraphNode findNode(Point pr){

	GraphNode nInit = null;

	int test =0;

	int hs = headList.size();
	for(int i=0;i<hs;i++){
	    nInit = (GraphNode)(headList.elementAt(i));
	    Triangle trian = nInit.getTriangle();
	    if( trian.contains(pr) ){
		//
		test = 1;
		//
		break;
	    }
	}
	//
	if (test==0){
	    System.err.println("#### Probleme: pas de triangle initial trouve");
	}
	//
	return searchRecNode(pr, nInit);
	
    } 


    
    /**
     * Recursive method which returns the triangle of the Delaunay triangulation (which has to be a leaf of the graph) containing the Point pr
     * @param pr the Point that is to be located
     * @param node the node where we look for "pr"
     * @return the child of "node" in which "pr" is located
     */
    private GraphNode searchRecNode(Point pr, GraphNode node){

	if ( (node.getChildList()).isEmpty() ){
	    return node;
	}
	else{
	    //
	    int test =0;
	    //
	    GraphNode n = null;
	    Vector list = node.getChildList();
	    int ls = list.size();
	    for(int i=0;i<ls;i++){
		n = (GraphNode)(list.elementAt(i));
		if((n.getTriangle()).contains(pr)){
		    //
		    test =1;
		    //
		    break;
		}	
	    }
	    //
	    if (test==0){
		System.err.println("#### Probleme: pas de triangle trouve avec la fonction recursive");
	    }
	    //
	    return searchRecNode(pr,n);
	}
    } 




    /**
     * Generates randomly and uniformly a point on the unit sphere. 
     * @return p the generated point
     */
    private Point pointGeneration(int nb){

	double phi = (generator.nextDouble())*(2.0*Math.PI);

	double z = ( ((generator.nextDouble())*2.0) - 1.0);
	double theta = Math.acos(z);


	Point p = new Point (1,theta,phi,1);
	p.setName((""+nb));

	return p;

    }

    
    /**
     * Displays all the leaves of the graph
     */
    public void printLeaf(){

	int ls = leafList.size();
	for(int i=0;i<ls;i++){
	        
	    Triangle tr = ((GraphNode)(leafList.elementAt(i))).getTriangle();
	    tr.printCoordinates();
	    System.out.println("---------------------------------------------------------------");
	}

    }


    /**
     * Displays the nodes representing the top of the graph
     */
    public void printHead(){

	int hs = headList.size();
	for(int i=0;i<hs;i++){
	        
	    Triangle tr = ((GraphNode)(headList.elementAt(i))).getTriangle();
	    tr.printCoordinates();
	    System.out.println("---------------------------------------------------------------");
	}

    }


    /**
     * Getter for the leaf list
     * @return leafList Vector containing all the leaves
     */
    public Vector getLeafList(){
	return leafList;
    }


}
