// Point class
// Creation : 18/11/2008
// Last modification : 20/01/2009


import java.lang.Math.*;
import java.util.Vector;


/** 
 * Class representing a point
 */
public class Point {


    private double x,y,z;  
    private double rayon,theta,phi;  

    private String name = ""; 



    /**
     * Constructor of a point
     * @param q1 Radius or X-coordinate
     * @param q2 Colatitude (between 0 and pi) or Y-coordinate
     * @param q3 Longitude (between 0 and 2*pi) or Z-coordinate
     * @param sphere Integer (1 if spherical coordinates, 0 if cartesian coordinates)
     */
    public Point (double q1, double q2, double q3, int sphere) {

	if (sphere == 1){
	    if (q1 < 0){

		System.out.println("Error in Point.java: radius < 0");
		System.exit(1);
	    }
	    rayon = q1;
	    theta = Math.acos(Math.cos(q2));
	    phi = (((q3 % (2*Math.PI)) + 2*Math.PI) % (2*Math.PI));
	
	    double xrec = rayon*Math.cos(phi)*Math.sin(theta);
	    double yrec = rayon*Math.sin(phi)*Math.sin(theta);
	    double zrec = rayon*Math.cos(theta);

	    x = xrec;
	    y = yrec;
	    z = zrec;
	}
	else if (sphere == 0){
	    x = q1;
	    y = q2;
	    z = q3;
	    rayon = Math.sqrt(x*x + y*y + z*z);
	    if (rayon < 1e-10){
		theta = 0;
		phi = 0;
	    }
	    else {
		theta = Math.acos(z/rayon);
		if ((theta < 1E-10)||(Math.abs((theta - Math.PI))< 1E-10)){
		    phi = 0;
		}
		else {
		    phi = Math.atan(y/x);
		    if (x<0){
			phi = phi + Math.PI;
		    }
		    phi = ((phi % (2*Math.PI) + 2*Math.PI)) % (2*Math.PI);
		}
	    }
	}
	else {
	    System.out.println("Error : Bad Point creation in Point.java");
	    System.exit(1);
	}
    }

    

    /** 
     * Return the cartesian coordinates of the point
     * @return Vector (X,Y,Z)
     */
    public Vector getCartCoordinates() {
	Vector pos = new Vector(3);
	Double q1 = new Double(x);
	Double q2 = new Double(y);
	Double q3 = new Double(z);
	pos.addElement(q1);
	pos.addElement(q2);
	pos.addElement(q3);
	return pos;
    }



    /** 
     * Return the spherical coordinates of the point
     * @return Vector (r,theta,phi)
     */
    public Vector getSphereCoordinates() {
	Vector pos = new Vector(3);
	Double q1 = new Double(rayon);
	Double q2 = new Double(theta);
	Double q3 = new Double(phi);
	pos.addElement(q1);
	pos.addElement(q2);
	pos.addElement(q3);
	return pos;
    }
	 

 
    /** 
     * Display the point coordinates
     */
    public void printPosition() {
	System.out.println("Cartesiennes      x= " + this.getCartCoordinates().elementAt(0) +  " y= " + this.getCartCoordinates().elementAt(1) + " z= " + this.getCartCoordinates().elementAt(2) + "\n" +  "Spheriques        r= " + this.getSphereCoordinates().elementAt(0) +  " theta= " + this.getSphereCoordinates().elementAt(1) + " phi= " + this.getSphereCoordinates().elementAt(2) + "\n");
    }



    /** 
     * Display the point cartesian coordinates
     */
    public void printCartCoordinates() {
	System.out.println("Cartesiennes      x= " + this.getCartCoordinates().elementAt(0) +  " y= " + this.getCartCoordinates().elementAt(1) + " z= " + this.getCartCoordinates().elementAt(2) + "\n");
    }



    /**
     * Allocates a name to the point. Useful for testing.
     * @param n the name allocated to the point 
     */
    public void setName(String n){
	name = n;
    }



    /**
     * Getter for the associated name
     */
    public String getName(){
	return name;
    }



    /** 
     * Return the image of the point by a rotation around an axis.
     * @param theta Rotation angle (rads)
     * @param axis Rotation axis (1 for X-axis, 2 for Y-axis, 3 for Z-axis)
     * @return Rotated point
     */
    public Point rotate(double theta, int axis) {
	if ((axis > 3)||(axis == 0)) {
	    System.out.println("Erreur dans methode POint.rotate : axe mal dfini");
	    System.exit(1);
	}

	double x = this.x;
	double y = this.y;
	double z = this.z;

	double xr;
	double yr;
	double zr;
	
	if (axis == 1) {
	    xr = x;
	    yr = y*Math.cos(theta) - z*Math.sin(theta);
	    zr = y*Math.sin(theta) + z*Math.cos(theta);
	}
	else if (axis == 2) {
	    yr = y;
	    xr = x*Math.cos(theta) + z*Math.sin(theta);
	    zr = -x*Math.sin(theta) + z*Math.cos(theta);
	}
	else {
	    zr = z;
	    xr = x*Math.cos(theta) - y*Math.sin(theta);
	    yr = x*Math.sin(theta) + y*Math.cos(theta);
	}
	
	Point P = new Point(xr,yr,zr,0);
	return P;
    }



    /** 
     * Return the middle of 2 points
     * @param Q Second point
     * @return Middle  
     */
    public Point middle(Point Q) {
	double xp = this.x;
	double yp = this.y;
	double zp = this.z;

	double xq = Q.x;
	double yq = Q.y;
	double zq = Q.z;

	double a = (xp + xq) / 2.0;
	double b = (yp + yq) / 2.0;
	double c = (zp + zq) / 2.0; 

	Point M = new Point(a,b,c,0);
	return M;
    }



    /** 
     * Return the equation of a segment bissector. The segment and its bissector are the same plan given by the perpendicular vector n
     * @param Q Second point giving the segment PQ
     * @param n Vector giving the equation of the plan
     * @return Vector containing A,B,C,D,A',B',C',D' where the equation is  Ax + By + Cz + D = 0 et  A'x + B'y + C'z + D' = 0
     */
    public Vector bissector(Point Q, Vector n) {
	
	Vect PQ = new Vect(this,Q);

	double A = ((Double)(n.elementAt(0))).doubleValue();
	double B = ((Double)(n.elementAt(1))).doubleValue();
	double C = ((Double)(n.elementAt(2))).doubleValue();
	double D = ((Double)(n.elementAt(3))).doubleValue();

	PQ = PQ.normalize();
	Vector W = PQ.getCoordinates();
	
	double A_ =  ((Double)(W.elementAt(0))).doubleValue();
	double B_ =  ((Double)(W.elementAt(1))).doubleValue();
	double C_ =  ((Double)(W.elementAt(2))).doubleValue();
	
	// reste a fixer D' en disant que le milieu de PQ est dans le plan
	
	Point M = this.middle(Q);

	Vector Mc = M.getCartCoordinates();

	double xc =  ((Double)(Mc.elementAt(0))).doubleValue();
	double yc =  ((Double)(Mc.elementAt(1))).doubleValue();
	double zc =  ((Double)(Mc.elementAt(2))).doubleValue();
	
	double D_ = -(A_*xc + B_*yc + C_*zc);

	Vector R = new Vector(8);
	R.addElement(new Double(A));
	R.addElement(new Double(B));
	R.addElement(new Double(C));
	R.addElement(new Double(D));
	R.addElement(new Double(A_));
	R.addElement(new Double(B_));
	R.addElement(new Double(C_));
	R.addElement(new Double(D_));
	    
      	return R;
    }



    /** 
     * Return the intersection of two segment bisectors given by 3 points in the plane containing these 3 points
     * @param Q Second point giving the first segment PQ
     * @param R Third point giving the second segment PR
     * @return Intersection point  
     */

    public Point intersec(Point Q, Point R) {
	Triangle T = new Triangle(this,Q,R);
	Vector P = T.plan();

	Vector Z1 = this.bissector(Q,P);
	Vector Z2 = this.bissector(R,P);

	double A0 = ((Double)(Z1.elementAt(0))).doubleValue();
	double B0 = ((Double)(Z1.elementAt(1))).doubleValue();
	double C0 = ((Double)(Z1.elementAt(2))).doubleValue();
	double D0 = ((Double)(Z1.elementAt(3))).doubleValue();

	double A1 = ((Double)(Z1.elementAt(4))).doubleValue();
	double B1 = ((Double)(Z1.elementAt(5))).doubleValue();
	double C1 = ((Double)(Z1.elementAt(6))).doubleValue();
	double D1 = ((Double)(Z1.elementAt(7))).doubleValue();

	double A2 = ((Double)(Z2.elementAt(4))).doubleValue();
	double B2 = ((Double)(Z2.elementAt(5))).doubleValue();
	double C2 = ((Double)(Z2.elementAt(6))).doubleValue();
	double D2 = ((Double)(Z2.elementAt(7))).doubleValue();

	// Formation du systeme matriciel

	Matrix A = new Matrix(3,3);
	
	A.set(0,0,A0);
	A.set(0,1,B0);
	A.set(0,2,C0);

	A.set(1,0,A1);
	A.set(1,1,B1);
	A.set(1,2,C1);

	A.set(2,0,A2);
	A.set(2,1,B2);
	A.set(2,2,C2);

	Matrix B = new Matrix(3,1);

	B.set(0,0,-D0);
	B.set(1,0,-D1);
	B.set(2,0,-D2);

	// Resolution de A*X = B
	
	LUDecomposition L = new LUDecomposition(A);

	Matrix X = L.solve(B);
	
	double x = X.get(0,0);
	double y = X.get(1,0);
	double z = X.get(2,0);

	Point I = new Point(x,y,z,0);

	return I;	
    }


    /** 
     * Returns the euclidian distance between 2 points
     * @param point Point 
     * @return Distance between the 2 points
     */
    public double eucliDist(Point point) {
	Vect V = new Vect(this,point);
	double d = V.norm();
	return d;
    }



}
